from coverage_strategies.src.Entities import Slot, Strategy


class CircleOutsideFromIo_Strategy(Strategy):
    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        """
        This function returns the coverage steps, choosing to start from the initial position of the opponent, io, and from
        there cover the world circling out.
        This function perform spiraling out simply: doesn't take steps in alternate sizes, and limiting the available slots
        to be in range
        :param self: the agent
        :param board_size:
        :param agent_o:
        :return:
        """

        fixed_io = Slot(agent_o.InitPosX, agent_o.InitPosY)
        self.steps.extend(
            Strategy.go_from_a_to_b(
                Slot(agent_r.InitPosX, agent_r.InitPosY),
                fixed_io))

        # cover the world, circling from this position out.
        current_slot = self.steps[-1]
        step_size = 1
        counter = 1

        b = [[0 for i in xrange(board_size)] for j in xrange(board_size)]
        b[current_slot.row][current_slot.col] = 1

        while sum(map(sum, b)) < board_size*board_size:
            # go horizontally right then vertically up
            for dir in ['r','u']:
                for _ in xrange(step_size):
                    current_slot = current_slot.go(dir)
                    counter += 1
                    if 0 <= current_slot.row < board_size and 0 <= current_slot.col < board_size:
                        self.steps.append(current_slot)
                        b[current_slot.row][current_slot.col] = 1

            step_size += 1
            # go horizontally left then vertically down
            for dir in ['l','d']:
                for _ in xrange(step_size):
                    current_slot = current_slot.go(dir)
                    counter += 1
                    if 0 <= current_slot.row < board_size and 0 <= current_slot.col < board_size:
                        self.steps.append(current_slot)
                        b[current_slot.row][current_slot.col] = 1
            step_size += 1
        return self.steps



from coverage_strategies.src.Entities import Strategy, Slot


class HorizontalCircularCoverage_Strategy(Strategy):
    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        next_slot = Slot(agent_r.InitPosX, agent_r.InitPosY)
        flag = (agent_r.InitPosX == board_size - 1)
        counter = 0

        # print "init pos {},{}: ".format(agent_r.InitPosX, agent_r.InitPosY)

        while True:
            counter += 1
            if counter > board_size * board_size:
                break

            self.steps.append(next_slot)
            # in the middle of moving from bottom rpw to top row
            if flag:
                if next_slot.col == board_size - 1:
                    flag = False
                    next_slot = next_slot.go_north()
                else:
                    next_slot = next_slot.go_east()
                if next_slot == Slot(agent_r.InitPosX, agent_r.InitPosY):
                    break
                continue
            # check if in last position, and start moving from last row to top row
            elif next_slot.row == board_size - 1 - 1 and next_slot.col == 0:
                flag = True
                next_slot = next_slot.go_south()

                if next_slot == Slot(agent_r.InitPosX, agent_r.InitPosY):
                    break
                continue
            # update next slot
            elif next_slot.col % 2 != 0:
                if next_slot.row == 0:
                    next_slot = next_slot.go_west()
                else:
                    next_slot = next_slot.go_north()
            else:
                if next_slot.row == board_size - 1 - 1:
                    next_slot = next_slot.go_west()
                else:
                    next_slot = next_slot.go_south()
                    if next_slot.row > board_size - 1:
                        print("+1")

            if next_slot == Slot(agent_r.InitPosX, agent_r.InitPosY):
                break

        return self.steps



from coverage_strategies.src.Entities import Strategy, Slot


class CircleOutsideFromBoardCenter_Strategy(Strategy):
    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        """
            This function return the agent_r steps, when deciding to cover the world, spiraling from inside to outside.
            Note: this coverage method is not optimal!
            :param agent_r: the agent covering the world
            :param board_size: self explanatory
            :return: list of steps
            """

        self.steps.extend(self.go_from_a_to_b(Slot(agent_r.InitPosX, agent_r.InitPosY), Slot(board_size/2, board_size/2)))

        next_slot = self.steps[-1]
        # next_slot = Slot(36,6)

        # start by going toward the center
        if next_slot.row < board_size / 2:
            while next_slot.row < board_size / 2:
                next_slot = next_slot.go_south()
                self.steps.append(next_slot)
                continue
        else:
            while next_slot.row > board_size / 2:
                next_slot = next_slot.go_north()
                self.steps.append(next_slot)
                continue

        if next_slot.col < board_size / 2:
            while next_slot.col < board_size / 2:
                next_slot = next_slot.go_east()
                self.steps.append(next_slot)
                continue
        else:
            while next_slot.col > board_size / 2:
                next_slot = next_slot.go_west()
                self.steps.append(next_slot)
                continue

        # after reaching the center, start covering, counter clockwise
        circ = 0
        counter = 1
        while circ < board_size:
            circ += 1
            if circ < board_size:
                for _ in range(circ):
                    next_slot = next_slot.go_west()
                    self.steps.append(next_slot)
                    counter += 1
            if circ < board_size:
                for _ in range(circ):
                    next_slot = next_slot.go_north()
                    self.steps.append(next_slot)
                    counter += 1

            circ += 1
            if circ < board_size:
                for _ in range(circ):
                    next_slot = next_slot.go_east()
                    self.steps.append(next_slot)
                    counter += 1
            if circ < board_size:
                for _ in range(circ):
                    next_slot = next_slot.go_south()
                    self.steps.append(next_slot)
                    counter += 1

        for step in self.steps:
            if step.row > board_size-1 or step.col > board_size-1 or step.row < 0 or step.col < 0:
                print("Error while circling outside from board center: reached unavailable position")
                break
        return self.steps



from coverage_strategies.src.Entities import Slot, Strategy


class VerticalCoverageFromCornerFarthestFromIo_Strategy(Strategy):
    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        """
            This function returns the coverage self.steps,when covering knowing io, and start covering vertically from the farthest
            corner.
            :param self:
            :param board_size:
            :param agent_o:
            :return: the coverage self.steps.
            """
        assert agent_o is not None

        # go to the farthest corner
        self.steps.extend(Strategy.go_from_a_to_b(a=Slot(agent_r.InitPosX, agent_r.InitPosY),
                                                  b=Strategy.get_farthest_corner(
                                                      a=Slot(agent_o.InitPosX, agent_o.InitPosY),
                                                      board_size=board_size)))

        # from there, cover vertically
        current_slot = self.steps[-1]
        v_dir = 'u' if current_slot.row == board_size - 1 else 'd'
        h_dir = 'l' if current_slot.col == board_size - 1 else 'r'
        counter = 1
        while counter <= board_size * board_size:
            if v_dir == 'u':
                while current_slot.row > 0:
                    current_slot = current_slot.go_north()
                    self.steps.append(current_slot)
                    counter += 1

                if counter == board_size * board_size:
                    break

                current_slot = current_slot.go_west() if h_dir == 'l' else current_slot.go_east()
                v_dir = 'd'
                self.steps.append(current_slot)
                counter += 1
                continue
            else:
                while current_slot.row < board_size - 1:
                    current_slot = current_slot.go_south()
                    self.steps.append(current_slot)
                    counter += 1

                if counter == board_size * board_size:
                    break

                current_slot = current_slot.go_west() if h_dir == 'l' else current_slot.go_east()
                v_dir = 'u'
                self.steps.append(current_slot)
                counter += 1
                continue

        return self.steps



from coverage_strategies.src.Entities import Slot, Strategy
from abc import abstractmethod

class CoverByQuarters_Strategy(Strategy):
    @abstractmethod
    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        # This coverage strategy covers first the top left, then top right, then bottom left, then bottom right quarters of
        # the area.
        # While building this function, we assumed 100X100 dimensions

        next_slot = Slot(agent_r.InitPosX, agent_r.InitPosY)
        # flag = (agent_r.InitPosY == boardSizeY - 1 and not (agent_r.InitPosX == 0))

        steps = []
        counter = 0

        while True:
            counter += 1

            if counter > 100000:
                print("Something is wrong, counter is too big!")
                print(steps)
                break

            if counter > 1 and (next_slot.row, next_slot.col) == (agent_r.InitPosX, agent_r.InitPosY):
                break

            steps.append(next_slot)

            # TL Quarter
            if 0 <= next_slot.row < board_size / 2 and 0 <= next_slot.col < board_size / 2:
                if (next_slot.row, next_slot.col) == (board_size / 2 - 1, board_size / 2 - 1):
                    next_slot = next_slot.go_east()
                    continue
                if next_slot.col == 0:
                    next_slot = next_slot.go_north() if next_slot.row > 0 else next_slot.go_east()
                    continue
                if next_slot.row % 2 == 0 and next_slot.row < board_size / 2 - 2:  # An even line, not the last one
                    next_slot = next_slot.go_east() if not next_slot.col == board_size / 2 - 1 else next_slot.go_south()
                    continue
                elif next_slot.row % 2 != 0 and not next_slot.row == board_size / 2 - 1:  # An odd line, not before last
                    next_slot = next_slot.go_west() if not next_slot.col == 1 else next_slot.go_south()
                    continue
                elif next_slot.row % 2 == 0 and next_slot.row == board_size / 2 - 2:  # An even line, the last one
                    next_slot = next_slot.go_south() if next_slot.col % 2 != 0 else next_slot.go_east()
                elif next_slot.row % 2 != 0 and next_slot.row == board_size / 2 - 1:  # An odd line, last line
                    next_slot = next_slot.go_east() if next_slot.col % 2 != 0 else next_slot.go_north()
                else:
                    print("TL: Error occurred! Should not reach here!")
                continue

            # TR Quarter
            elif 0 <= next_slot.row < board_size / 2 and board_size / 2 <= next_slot.col < board_size:
                if (next_slot.row, next_slot.col) == (board_size / 2 - 1, board_size - 1):
                    next_slot = next_slot.go_south()
                    continue
                elif next_slot.col % 2 == 0:
                    next_slot = next_slot.go_north() if next_slot.row > 0 else next_slot.go_east()
                    continue
                elif next_slot.col % 2 != 0:
                    next_slot = next_slot.go_south() if next_slot.row < board_size / 2 - 1 else next_slot.go_east()
                    continue
                else:
                    print("TR: Error occurred! Should not reach here!")
                continue

            # BL Quarter
            elif board_size / 2 <= next_slot.row < board_size and 0 <= next_slot.col < board_size / 2:
                if (next_slot.row, next_slot.col) == (board_size / 2, 0):  # last cell of quarter
                    next_slot = next_slot.go_north()
                    continue
                elif next_slot.col % 2 == 0:  # an even column
                    next_slot = next_slot.go_north() if next_slot.row > board_size / 2 else next_slot.go_west()
                    continue
                elif next_slot.col % 2 != 0:  # An odd line
                    next_slot = next_slot.go_south() if next_slot.row < board_size - 1 else next_slot.go_west()
                    continue
                else:
                    print("BL: Error occurred! Should not reach here!")
                continue

            # BR Quarter
            else:
                if (next_slot.row, next_slot.col) == (board_size / 2, board_size / 2):  # last cell of quarter
                    next_slot = next_slot.go_west()
                    continue
                elif next_slot.col == board_size - 1:
                    next_slot = next_slot.go_south() if not next_slot.row == board_size - 1 else next_slot.go_west()
                    continue
                elif next_slot.row % 2 != 0 and not next_slot.row == board_size / 2 + 1:  # and odd line, not before last
                    next_slot = next_slot.go_west() if next_slot.col > board_size / 2 else next_slot.go_north()
                    continue
                elif next_slot.row % 2 != 0 and next_slot.row == board_size / 2 + 1:  # and odd line, DO before last
                    next_slot = next_slot.go_north() if next_slot.col % 2 == 0 else next_slot.go_west()
                    continue
                elif next_slot.row % 2 == 0 and not next_slot.row == board_size / 2:  # an even line, not last one
                    next_slot = next_slot.go_east() if next_slot.col < board_size - 2 else  next_slot.go_north()
                    continue
                elif next_slot.row % 2 == 0 and next_slot.row == board_size / 2:  # an even line, INDEED last one
                    next_slot = next_slot.go_west() if next_slot.col % 2 == 0 else next_slot.go_south()
                    continue
                else:
                    print("BR: Error occurred! Should not reach here!")
                continue

        return steps



from random import shuffle, random
from math import floor
from coverage_strategies.src import Entities


def print_graph(edges, i_o, figure_label=""):
    pass
    # print("Function is not available in this version!")
    # plt.figure()
    # g = nx.DiGraph()
    # g.add_edges_from(edges)
    # positions = {}
    # color_map = []
    # for n in g.nodes():
    #     positions[n] = (n.row, n.col)
    #     if n == edges[0][0]:
    #         color_map.append('green')
    #     elif n == i_o:
    #         color_map.append('blue')
    #     else:
    #         color_map.append('red')
    #     # color_map.append('green' if n == edges[0][0] else 'red')
    #
    # f = plt.figure()
    # nx.draw(G=g, pos=positions, node_size=20, font_size=8, node_color=color_map, ax=f.add_subplot(111))
    #
    # if not os.path.exists(figure_label):
    #     os.makedirs(figure_label)
    #
    # f.savefig(figure_label + "/max_path.png", bbox_inches='tight')
    # plt.close('all')
    #
    # # plt.axis([-1, 33, -1, 33], 'on')
    # # plt.grid(True)
    #
    # plt.show()


def create_graph(edge_list):
    graph = {}
    for e1, e2 in edge_list:
        graph.setdefault(e1, []).append(e2)
        graph.setdefault(e2, []).append(e1)
    return graph


def get_shallow_graph(original_edges):
    shallow_graph = {}
    for e1, e2 in original_edges:
        shallow_e1 = Entities.Slot(floor(e1.row / 2.0), floor(e1.col / 2.0))
        shallow_e2 = Entities.Slot(floor(e2.row / 2.0), floor(e2.col / 2.0))

        if (shallow_e1.row, shallow_e1.col) == (shallow_e2.row, shallow_e2.col):
            continue

        if shallow_e1 not in shallow_graph.keys() or shallow_e2 not in shallow_graph[shallow_e1]:
            shallow_graph.setdefault(shallow_e1, []).append(shallow_e2)

        if shallow_e2 not in shallow_graph.keys() or shallow_e1 not in shallow_graph[shallow_e2]:
            shallow_graph.setdefault(shallow_e2, []).append(shallow_e1)

    return shallow_graph


# Prim's
def mst(start, graph):
    closed = set()
    edges = []
    q = [(start, start)]
    while q:
        # randomize
        shuffle(q)

        v1, v2 = q.pop()
        if v2 in closed:
            continue
        closed.add(v2)
        edges.append((v1, v2))
        for v in graph[v2]:
            if v in graph:
                q.append((v2, v))
    del edges[0]
    assert len(edges) == len(graph)-1
    return edges


def create_covering_path(mst_edges_shallow_graph, initial_slot):
    covering_path = []
    origin_slot = initial_slot
    slot = origin_slot
    counter = 0
    while True:
        counter += 1
        if counter > 100000:
            print("ERROR! Probably hit infinite loop while creating the covering path")
            return

        covering_path.append(slot)
        shallow_slot = Entities.Slot(floor(slot.row / 2.0), floor(slot.col / 2.0))
        # find to where to go next, depend on the mst edges.
        # Check how much and which corners are in the mst group, then update slot accordingly

        has_downward_edge = (shallow_slot, shallow_slot.go_south()) in mst_edges_shallow_graph or \
                            (shallow_slot.go_south(), shallow_slot) in mst_edges_shallow_graph
        has_rightward_edge = (shallow_slot, shallow_slot.go_east()) in mst_edges_shallow_graph or \
                             (shallow_slot.go_east(), shallow_slot) in mst_edges_shallow_graph
        has_leftward_edge = (shallow_slot, shallow_slot.go_west()) in mst_edges_shallow_graph or \
                            (shallow_slot.go_west(), shallow_slot) in mst_edges_shallow_graph
        has_upward_edge = (shallow_slot, shallow_slot.go_north()) in mst_edges_shallow_graph or \
                          (shallow_slot.go_north(), shallow_slot) in mst_edges_shallow_graph

        bl_corner_in_mst = False
        br_corner_in_mst = False
        ul_corner_in_mst = False
        ur_corner_in_mst = False

        if slot.row % 2 == 0 and slot.col % 2 == 0:
            if has_downward_edge or has_rightward_edge:
                br_corner_in_mst = True
            if has_upward_edge:
                br_corner_in_mst = True
                ur_corner_in_mst = True
            if has_leftward_edge:
                bl_corner_in_mst = True
                br_corner_in_mst = True
        elif slot.row % 2 == 0 and slot.col % 2 != 0:
            if has_downward_edge or has_leftward_edge:
                bl_corner_in_mst = True
            if has_upward_edge:
                bl_corner_in_mst = True
                ul_corner_in_mst = True
            if has_rightward_edge:
                bl_corner_in_mst = True
                br_corner_in_mst = True
        elif slot.row % 2 != 0 and slot.col % 2 == 0:
            if has_rightward_edge or has_upward_edge:
                ur_corner_in_mst = True
            if has_downward_edge:
                br_corner_in_mst = True
                ur_corner_in_mst = True
            if has_leftward_edge:
                ur_corner_in_mst = True
                ul_corner_in_mst = True
        elif slot.row % 2 != 0 and slot.col % 2 != 0:
            if has_leftward_edge or has_upward_edge:
                ul_corner_in_mst = True
            if has_downward_edge:
                bl_corner_in_mst = True
                ul_corner_in_mst = True
            if has_rightward_edge:
                ul_corner_in_mst = True
                ur_corner_in_mst = True

        last_slot = covering_path[len(covering_path)-2]

        # check o see if only one vertex was in the mst
        if br_corner_in_mst and not (ur_corner_in_mst or bl_corner_in_mst or ul_corner_in_mst):
            if slot.go_south() == last_slot or slot == last_slot:
                slot = slot.go_east()
            elif slot.go_east() == last_slot:
                slot = slot.go_south()
        elif ur_corner_in_mst and not (br_corner_in_mst or bl_corner_in_mst or ul_corner_in_mst):
            if slot.go_east() == last_slot or slot == last_slot:
                slot = slot.go_north()
            elif slot.go_north() == last_slot:
                slot = slot.go_east()
        elif bl_corner_in_mst and not (ul_corner_in_mst or ur_corner_in_mst or br_corner_in_mst):
            if slot.go_west() == last_slot or slot == initial_slot:
                slot = slot.go_south()
            elif slot.go_south() == last_slot:
                slot = slot.go_west()
        elif ul_corner_in_mst and not (bl_corner_in_mst or br_corner_in_mst or ur_corner_in_mst):
            if slot.go_north() == last_slot or slot == initial_slot:
                slot = slot.go_west()
            elif slot.go_west() == last_slot:
                slot = slot.go_north()
        # check to see if exactly two vertices are in the mst
        elif bl_corner_in_mst and br_corner_in_mst and not (ul_corner_in_mst or ur_corner_in_mst):
            if slot.go_west() == last_slot or slot == initial_slot:
                slot = slot.go_east()
            elif slot.go_east() == last_slot:
                slot = slot.go_west()
            else:
                print("error 1")
        elif ul_corner_in_mst and ur_corner_in_mst and not (bl_corner_in_mst or br_corner_in_mst):
            if slot.go_east() == last_slot or slot == initial_slot:
                slot = slot.go_west()
            elif slot.go_west() == last_slot:
                slot = slot.go_east()
        elif br_corner_in_mst and ur_corner_in_mst and not (bl_corner_in_mst or ul_corner_in_mst):
            if slot.go_south() == last_slot or slot == initial_slot:
                slot = slot.go_north()
            elif slot.go_north() == last_slot:
                slot = slot.go_south()
        elif bl_corner_in_mst and ul_corner_in_mst and not (br_corner_in_mst or ur_corner_in_mst):
            if slot.go_north() == last_slot or slot == initial_slot:
                slot = slot.go_south()
            elif slot.go_south() == last_slot:
                slot = slot.go_north()
        # check for exactly 3 vertices
        elif br_corner_in_mst and bl_corner_in_mst and ul_corner_in_mst and not ur_corner_in_mst:
            if slot.go_north() == last_slot or slot == initial_slot:
                slot = slot.go_east()
            elif slot.go_east() == last_slot:
                slot = slot.go_north()
        elif bl_corner_in_mst and ul_corner_in_mst and ur_corner_in_mst and not br_corner_in_mst:
            if slot.go_south() == last_slot or slot == initial_slot:
                slot = slot.go_east()
            elif slot.go_east() == last_slot:
                slot = slot.go_south()
        elif ul_corner_in_mst and ur_corner_in_mst and br_corner_in_mst and not bl_corner_in_mst:
            if slot.go_west() == last_slot or slot == initial_slot:
                slot = slot.go_south()
            elif slot.go_south() == last_slot:
                slot = slot.go_west()
        elif ur_corner_in_mst and br_corner_in_mst and bl_corner_in_mst and not ul_corner_in_mst:
            if slot.go_north() == last_slot or slot == initial_slot:
                slot = slot.go_west()
            elif slot.go_west() == last_slot:
                slot = slot.go_north()
        else:
            print("error has occurred!")

        if slot == origin_slot:
            break

    # flip path direction half of the time
    if random() < 0.5:
        covering_path = flip_path(covering_path)

    return covering_path

def flip_path(path):
    return_path = []
    return_path.append(path[0])
    return_path.extend([path[i] for i in range(len(path)-1,0,-1)])
    return return_path

def is_slot_shallow_obstacle(slot, obstacles):
    return any([(int(slot.row / 2) == int(s.row / 2) and int(slot.col / 2) == int(s.col / 2)) for s in obstacles])

def get_edges_for_full_graph(width, height, obstacles):
    edges = [(Entities.Slot(row, col), Entities.Slot(row, col).go_south())
             for row in range(height) for col in range(width)
             if not (is_slot_shallow_obstacle(Entities.Slot(row, col), obstacles)
                     or is_slot_shallow_obstacle(Entities.Slot(row, col).go_south(), obstacles))
             ]
    edges.extend([(Entities.Slot(row, col), Entities.Slot(row, col).go_east())
             for row in range(height) for col in range(width)
             if not (is_slot_shallow_obstacle(Entities.Slot(row, col), obstacles)
                     or is_slot_shallow_obstacle(Entities.Slot(row, col).go_east(), obstacles))
             ])
    inbounds = [e for e in edges if e[0].row < width and e[1].row < width and e[0].col < height and e[1].col < height ]
    return inbounds


def get_random_coverage_strategy(size, i_r, i_o=None, print_mst=False, figure_label="", obstacles=[]):
    edges = get_edges_for_full_graph(size, size, obstacles=obstacles)

    # remove edges inside, or connected to, obstacles

    shallow_graph = get_shallow_graph(edges)
    shallow_init_pos = Entities.Slot(floor(i_r.row / 2.0), floor(i_r.col / 2.0))
    mst_edges_shallow_graph = mst(shallow_init_pos, shallow_graph)
    covering_path = create_covering_path(mst_edges_shallow_graph, i_r)
    if print_mst:
        covering_path_edges = []
        for i in range(0, len(covering_path)-1):
            covering_path_edges.append((covering_path[i], covering_path[i+1]))
        print_graph(covering_path_edges, i_o, figure_label=figure_label)
    return covering_path


def display_path(covering_path):
    covering_path_edges = []
    for i in range(0, len(covering_path) - 1):
        covering_path_edges.append((covering_path[i], covering_path[i + 1]))
    print_graph(covering_path_edges, covering_path_edges[0][0])


if __name__ == '__main__':
    get_random_coverage_strategy(32, Entities.Slot(16, 16), True)



from coverage_strategies.src.Entities import Strategy, Slot, Agent
import random

def sublist(lst1, lst2):
    return all([(x in lst2) for x in lst1])

class TrulyRandomStrategy(Strategy):
    @classmethod
    def get_steps(self, agent_r: Agent, board_size=50, agent_o:Agent=None):
        covered_cells = []
        current_slot = Slot(agent_r.InitPosX, agent_r.InitPosY)
        covered_cells.append(current_slot)
        all_slots = [Slot(i,j) for i in range(board_size) for j in range(board_size)]
        while not sublist(all_slots, covered_cells):
            # randomly select move
            new_slot = current_slot
            while not (0 <= new_slot.row < board_size and 0 <= new_slot.col < board_size) or new_slot == current_slot:
                new_slot = current_slot.go(random.choice(['r','l','u','d']))

            # update accordingly
            covered_cells.append(new_slot)
            current_slot = new_slot
            print(len(set(covered_cells))/(board_size*board_size*1.0))
        print(covered_cells)
        print(len(covered_cells))
        exit(2)
        return covered_cells




from coverage_strategies.src.Entities import Slot, Strategy
from abc import abstractmethod

class VerticalNonCircularCoverage_Strategy(Strategy):
    @abstractmethod
    def get_steps(self, agent_r, board_size = 50, agent_o = None):
        """
            Returns a non-circular vertical coverage, starting from agent_r's initial position to top-right position, then cover
            all cells from top-right to bottom-left, without repeating cells (there are some assumption regarding the initial
            position
            :param agent_r: the given agent_r
            :param board_size: board's width
            :param board_size: board's height
            :return: a set of steps covering the whole board
            """
        steps = []
        next_slot = Slot(agent_r.InitPosX, agent_r.InitPosY)
        turning_slot = Slot(agent_r.InitPosX, agent_r.InitPosY)
        reaching_to_farthest_corner = True
        up_or_down = 'd'

        # assert call
        assert agent_r.InitPosX % 2 == 0
        while True:
            steps.append(next_slot)
            if len(steps) >= board_size * board_size:
                break

            # Check if we agent_r reached the farthest corner of the board
            if next_slot == Slot(board_size - 1, board_size - 1):
                reaching_to_farthest_corner = False

            if reaching_to_farthest_corner:
                if next_slot.row < board_size - 1:
                    next_slot = next_slot.go_south()
                    continue
                elif next_slot.col < board_size - 1:
                    next_slot = next_slot.go_east()
                    continue
            else:
                if next_slot.col > agent_r.InitPosY:
                    if next_slot.col % 2 != 0:
                        next_slot = next_slot.go_north() if next_slot.row > 0 else next_slot.go_west()
                    else:
                        next_slot = next_slot.go_south() if next_slot.row < board_size - 2 else next_slot.go_west()
                    continue
                else:
                    if up_or_down == 'd':
                        if next_slot.go_south() != turning_slot:
                            if next_slot.row < board_size - 1:
                                next_slot = next_slot.go_south()
                            else:
                                next_slot = next_slot.go_west()
                                up_or_down = 'u'
                            continue
                        else:
                            turning_slot = turning_slot.go_west().go_west().go_north().go_north()
                            steps.append(next_slot.go_west())
                            steps.append(next_slot.go_west().go_south())
                            next_slot = next_slot.go_west().go_south().go_south()
                            continue
                    else:
                        if next_slot != turning_slot:
                            if next_slot.row > 0:
                                next_slot = next_slot.go_north()
                            else:
                                next_slot = next_slot.go_west()
                                up_or_down = 'd'
                            continue
                        else:
                            steps.append(next_slot.go_east())
                            steps.append(next_slot.go_east().go_north())
                            next_slot = next_slot.go_east().go_north().go_north()
                            continue

        return steps[:board_size * board_size]